//---------------------------------------------------------------------------
//
// Evj        Model the evolution of DNA binding site recognition
//
// Module:    GeneSeqPanel, a Swing panel that displays a genetic sequence.
// Author:    Paul C. Anagnostopoulos
// Created:   6 April 2005
//
// This computer program fits the description in the U.S. Copyright Act
// of a "United States Government work." It was written as a part of
// my/our official duties as a Government employee. This means it cannot
// be copyrighted. The program is freely available for use without a
// copyright notice, and there are no restrictions on its use, now or
// subsequently. We retain no rights in the program. Please acknowledge
// the source of this code in any releases you make.
//
//---------------------------------------------------------------------------


import java.awt.*;
import java.awt.event.*;
import java.awt.font.*;
import java.lang.Math;
import java.util.Arrays;
import javax.swing.*;
import javax.swing.event.*;

//                      Gene Sequence Panel
//                      ---- -------- -----


public final class GeneSeqPanel extends JPanel implements NexusObserver
{
  private static final int gutter = 10;
  private static final int row0Baseline = 16;
  private static final int rowLeading = 100;
  private static final int baseLeading = 20;
  private static final int separatorWidth = 2;
  private static final int separatorHeight = baseLeading - 1;
  private static final int barOffset = 30;
  private static final int barLeading = 17;
  private static final int barHeight = 14;
  private static final int barShortfall = 5;
  private static final int splitOverset = 20;

  private static final Font titleFont  = new Font("Serif", Font.PLAIN, 22);
  private static final Font numberFont = new Font("SansSerif", Font.PLAIN, 12);
  private static final Font baseFont   = new Font("SansSerif", Font.BOLD, 17);
  private static final Font barFont    = new Font("SansSerif", Font.PLAIN, 13);

  private static final Color posWeightColor = Color.cyan;
  private static final Color negWeightColor = new Color(0.5f, 0.7f, 1.0f);
  private static final Color unmatchedSiteColor = Color.red;
  private static final Color matchedSiteColor = Color.green;
  private static final Color spuriousMatchColor = new Color(1.0f, 0.8f, 0.0f);

  private static final float minSaturation = 0.15f;
  private static final float maxSaturation = 0.8f - minSaturation;
  
  private JScrollPane scrollPane;
  private Nexus acceptedNexus = null;

  // The constructor saves the parent scroll pane.
  
  public GeneSeqPanel (JScrollPane sp)
  {
    super();
    scrollPane = sp;
  }

  // This public method implements the NexusObserver interface. It
  // simply saves the supplied nexus and repaints the display.
  
  public void acceptNexus (Nexus nexus)
  {
    acceptedNexus = nexus;
    repaint();
  }

//                      Paint Component
//                      ----- ---------


  private Graphics2D g2;
  private int baseTracking;
  private int basesPerRow;

  public void paintComponent (Graphics g)
  {
    super.paintComponent(g);

    // Pick up the nexus so we are impervious to nexuses coming in
    // while drawing. If it is null, forget it.
    
    Nexus nexus = acceptedNexus;
    if (nexus == null) return;
    g2 = (Graphics2D)g;

    int weightWidth = nexus.weightWidth;
    int siteWidth = nexus.siteWidth;

    int scrollWidth = scrollPane.getViewport().getExtentSize().width;
    int pageWidth = scrollWidth - 2*gutter;
    int pageHeight = scrollPane.getViewport().getExtentSize().height -
                     2*gutter;
    g2.translate(gutter, gutter);

    g2.setFont(baseFont);
    FontRenderContext frc = g2.getFontRenderContext();
    int baseSpan = (int)Math.round(
                     1.75 * baseFont.getStringBounds("G", frc).getWidth());

    int maxScore = (1 << 2 * weightWidth) / 2 * siteWidth;
    int scoreDigits = (int)Math.round(log10(maxScore) + 0.5);
    String template = "0A +" + "0000000000".substring(0, scoreDigits);
    g2.setFont(barFont);
    frc = g2.getFontRenderContext();
    int templateSpan = 3 +
                       (int)Math.round(
                         barFont.getStringBounds(template, frc).getWidth()) +
                       1 + barShortfall;

    baseTracking = Math.max(baseSpan,
                            Math.max(templateSpan / weightWidth,
                                     templateSpan / siteWidth));

    float widthRatio = 0;
    int weights;
    for (weights = 24; weights >= 8; weights -= 2) {
      widthRatio = (float)pageWidth / (baseTracking * weightWidth * weights);
      if (widthRatio >= 0.9) break;
    }
    basesPerRow = weights * weightWidth;

    int rows = (int)Math.round(
                 (double)nexus.chromosome.length / basesPerRow + 0.5);
    float heightRatio = (float)pageHeight / (rows * rowLeading);
    double scale = Math.max(Math.min(widthRatio, heightRatio), 0.9);
    if (scale < 1.0) g2.scale(scale, scale);
    setPreferredSize(new Dimension(scrollWidth,
                                   rows * rowLeading + 2*gutter));
    revalidate();

    drawChromosome(nexus);
    drawPositions(nexus);
    drawBars(nexus);
  }

//                      Draw Chromosome
//                      ---- ----------


  private void drawChromosome (Nexus nexus)
  {
    Paint paint = g2.getPaint();
    byte chromosome[] = nexus.chromosome;
    int width = nexus.siteWidth;

    g2.setFont(baseFont);
    int x = 0;
    int y = row0Baseline + baseLeading - rowLeading;
    for (int p = 0; p < chromosome.length; ++p) {
      if (p % basesPerRow == 0) { y += rowLeading; x = 0; }
      g2.setPaint(Base.color(chromosome[p]));
      g2.drawString(Base.letter(chromosome[p]), x, y);
      if (isSeparatorPos(nexus, p)) {
        g2.setPaint(paint);
        g2.fillRect(x + baseTracking - 3*separatorWidth,
                    y - separatorHeight + separatorWidth,
                    separatorWidth, separatorHeight);
      }
      x += baseTracking;
    }
    g2.setPaint(paint);
  }

  // This method determines whether p is a chrosome position that needs
  // a separator line drawn after it. Separator lines are drawn after
  // each group of four weights in the gene, and also just after the
  // last possible site position.
  
  private boolean isSeparatorPos (Nexus nexus, int p)
  {
    int stride = 4 * nexus.weightWidth;
    
    if (p < nexus.thresholdPos && p % stride == stride-1) return true;
    if (p == nexus.chromosome.length - nexus.siteWidth) return true;
    return false;
  }

//                      Draw Positions
//                      ---- ---------

  // This method draws a tick mark (plus sign) at every fifth position
  // of the chromosome. In addition, it includes the position number on
  // every other tick mark.
  
  private void drawPositions (Nexus nexus)
  {
    g2.setFont(numberFont);
    int x, y;

    for (int p = 5; p < nexus.chromosome.length; p += 5) {
      int mod = p % basesPerRow;
      if (mod == 0) {
        x = (basesPerRow - 1) * baseTracking;
        y = row0Baseline + (p / basesPerRow - 1) * rowLeading;
      } else {
        x = (mod - 1) * baseTracking;
        y = row0Baseline + p / basesPerRow * rowLeading;
      }
      g2.drawString("+", x+2, y);
      if (p % 10 == 0) g2.drawString(String.valueOf(p), x+10, y);
    }
  }

//                      Draw Bars
//                      ---- ----


  private int prevBar[] = new int[3];

  private void drawBars (Nexus nexus)
  {
    PosInfo info[] = nexus.posInfo;
    
    g2.setFont(barFont);

    Arrays.fill(prevBar, -1);
    for (int i = 0, wb = -1; i < info.length; ++i) {
      int pos = info[i].position;
      int valuation = info[i].weight;
      int level;
      for (level = 0; level < prevBar.length-1; ++level)
        if (prevBar[level] < pos) break;
      prevBar[level] = pos + info[i].width - 1;

      switch (info[i].type) {
      case PosInfo.GENE_WEIGHT:
        drawOneBar(pos, info[i].width, level,
                   weightBarColor(nexus, info[i].width, valuation),
                   String.valueOf(++wb / 4) + Base.letter(wb % 4),
                   valuation); break;

      case PosInfo.GENE_THRESHOLD:
        drawOneBar(pos, info[i].width, level,
                   weightBarColor(nexus, info[i].width, valuation),
                   "thr", valuation); break;

      case PosInfo.SITE_UNMATCHED:
        drawOneBar(pos, info[i].width, level,
                   siteBarColor(nexus, unmatchedSiteColor, valuation),
                   "", valuation); break;

      case PosInfo.SITE_MATCHED:
        drawOneBar(pos, info[i].width, level,
                   siteBarColor(nexus, matchedSiteColor, valuation),
                   "", valuation); break;

      case PosInfo.NONSITE_MATCHED:
        drawOneBar(pos, info[i].width, level,
                   siteBarColor(nexus, spuriousMatchColor, valuation),
                   "", valuation); break;

//-      default:
//-        assert false: info[i].type; break;
      }
    }
  }

  private void drawOneBar (int pos, int width, int level, Color color,
                           String prefix, int weight)
  {
    Paint paint = g2.getPaint();

    int col = pos % basesPerRow;
    int row = pos / basesPerRow;
    int x = col * baseTracking;
    int y = row0Baseline + row * rowLeading + barOffset + level * barLeading;
    g2.setPaint(color);
    if (col + width <= basesPerRow)
      g2.fillRect(x, y, baseTracking * width - barShortfall, barHeight);
    else
      drawSplitBar(x, y, col, width);
    g2.setPaint(paint);
    String label = prefix + "  " +
                   ((weight > 0) ? "+" : "") + String.valueOf(weight);
    g2.drawString(label, x+3, y+barHeight-3);
  }

  // This method draws a bar that needs to be split across two rows of
  // the display. The bar sticks out to the right of the first row a
  // bit.
  
  private void drawSplitBar (int x, int y, int col, int width)
  {
    g2.fillRect(x, y,
                baseTracking * (basesPerRow - col) + splitOverset,
                barHeight);
    g2.fillRect(0, y + rowLeading,
                baseTracking * (width - basesPerRow + col) - barShortfall,
                barHeight);
  }

  // This method calculates the alpha value for a weight or threshold
  // bar color. It ranges from minSaturation to maxSaturation depending
  // on where the value falls in the overall range of possible values.
  
  private Color weightBarColor (Nexus nexus, int width, int valuation)
  {
    int maxWeight = 1 << (2 * width - 1);
    float d, alpha;
    Color base;
  
    if (valuation < 0) {
      d = (float)-valuation / maxWeight;
      base = negWeightColor;
    } else {
      d = (float)valuation / maxWeight;
      base = posWeightColor;
    }
    alpha = minSaturation + maxSaturation * d;

    return new Color(base.getRed(), base.getGreen(), base.getBlue(),
                     (int)(alpha * 255));
  }
    
  // This method calculates the alpha value for a site bar color. It
  // ranges from minSaturation to maxSaturation depending on the
  // difference between the valuation of this site and the
  // minimum/maximum possible valuation of any site.
  
  private Color siteBarColor (Nexus nexus, Color base, int valuation)
  {
    float d, alpha;
    
    if (valuation < nexus.threshold) {
      d = (float)absdiff(nexus.threshold, valuation) /
                 absdiff(nexus.threshold, nexus.minValuation);
    } else {
      d = (float)absdiff(nexus.threshold, valuation) /
                 absdiff(nexus.threshold, nexus.maxValuation);
    }
    alpha = minSaturation + maxSaturation * d;
  
    return new Color(base.getRed(), base.getGreen(), base.getBlue(),
                     (int)(alpha * 255));
  }

//                      Utilities
//                      ---------


  // This function returns the absolute difference of its two
  // arguments.

  private int absdiff (int x, int y)
  {
    return Math.abs(x - y);
  }
  
  // This function returns the log base 10 of its argument.
  
  private double log10 (int x)
  {
    return Math.log(x) / Math.log(10);
  }
}
